/*
 * ptar - PROJET ESIAL RS 2011/2012
 * Copyright (C) 2011 Julien VAUBOURG / Valentin KELLER
 * Contact <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ptar.h"

/*
 * (corps des threads)
 * Extrait le fichier pointe par header_t->tarFile, dont le header a ete
 * extrait et synchronise le disque dur
 * @p_data Pointeur vers le header de thread correspondant au fichier
 */
void extract_files(void* p_data) {
	header_for_thread* header_t = (header_for_thread *) p_data;
	int fd_extractedFile, *errors_t, isDir;
	struct stat stat;
	DIR* dir;

	/* Allocation dynamique pour le pthread_exit */
	errors_t = malloc(sizeof(int));
	*errors_t = 0;

	/* Creation des repertoires qui composent le chemin du fichier, s'ils
	 * n'ont pas encore ete rencontres dans le tar */
	extract_dirs(header_t->attr.filename);

	/* Determination du type de fichier */
	switch(header_t->typeflag) {

		/* Fichier regulier */
		case FILE_REGULAR :
			*errors_t = extract_regular(header_t);
		break;

		/* Lien physique */
		case FILE_HARD :
			*errors_t = extract_link(header_t, LINK_HARD);
		break;

		/* Lien symbolique */
		case FILE_SYM :
			*errors_t = extract_link(header_t, LINK_SYM);
		break;

		/* Peripherique caracteres */
		case FILE_CDEVICE :
			*errors_t = extract_device(header_t, S_IFCHR);
		break;

		/* Peripherique blocs */
		case FILE_BDEVICE :
			*errors_t = extract_device(header_t, S_IFBLK);
		break;

		/* Repertoire */
		case FILE_DIR :
			*errors_t = extract_dir(header_t);
		break;

		/* Tube */
		case FILE_FIFO :
			*errors_t = extract_fifo(header_t);
		break;

		/* Inconnu */
		default :
			pfclose(header_t->tarFile);
			fprintf(stderr, "[FATAL] Type de fichier non reconnu");
			exit(EXIT_FAILURE);
	}

	/* Test des differents cas d'erreurs que peuvent partager tous les types
	 * de fichier a la creation */
	switch(errno) {
		case ENOSPC :
			pfclose(header_t->tarFile);
			fprintf(stderr, "[FATAL] Pas assez de place disponible sur le disque\n");
			exit(EXIT_FAILURE);

		case ENOMEM :
			pfclose(header_t->tarFile);
			fprintf(stderr, "[FATAL] Pas assez de memoire\n");
			exit(EXIT_FAILURE);

		case EROFS :
			pfclose(header_t->tarFile);
			fprintf(stderr, "[FATAL] Systeme de fichiers en read-only");
			exit(EXIT_FAILURE);

		case ENAMETOOLONG :
			fprintf(
				stderr,
				"[ERREUR] %s : nom de fichier trop long pour le systeme de fichiers\n",
				header_t->attr.fn_with_path
			);

			(*errors_t)++;
	}

	/* Definit les differents attributs du fichier, si ce ne sont pas des liens */
	if(header_t->typeflag != FILE_SYM && header_t->typeflag != FILE_HARD)
		*errors_t += set_attributes(&(header_t->attr));

	/* Test l'existence du fichier qui est cense venir de se faire extraire,
	 * sans dereferencer les liens symboliques */
	if(!lstat(header_t->attr.fn_with_path, &stat)) {

		/* Synchronisation du fichier, quelque soit son type */	
		fd_extractedFile = open(header_t->attr.fn_with_path, O_RDWR);

		/* Le fichier est un dossier, changement de procedure */
		if((isDir = errno == EISDIR)) {
			dir = opendir(header_t->attr.fn_with_path);
			fd_extractedFile = dirfd(dir);
		}

		/* Le fichier a bien pu etre cree par l'extraction */
		if(fd_extractedFile > 0) {

			/* Synchonisation */
			fsync(fd_extractedFile);

			/* Fermeture */
			if(dir)
				close(fd_extractedFile);
			else
				closedir(dir);

			if(header_t->userOptions->verbose)
				printf("[INFO] %s : extrait\n", header_t->attr.fn_with_path);
		}

	} else {
		fprintf(
			stderr,
			"[ERREUR] %s : n'a pas pu etre cree\n",
			header_t->attr.fn_with_path
		);

		(*errors_t)++;
	}

	/* Indique que l'emplacement du thread correspond a un thread mort,
	 * qu'il faut joindre */
	threadsAlive[header_t->iThread] = DEAD_THREAD;

	free(header_t);
	pthread_exit((void *) errors_t);
}
