/*
 * ptar - PROJET ESIAL RS 2011/2012
 * Copyright (C) 2011 Julien VAUBOURG / Valentin KELLER
 * Contact <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ptar.h"

/*
 * Determine si l'archive est gzippee ou non
 * @tarFile Archive a decompresser
 * @return Compression gzip detectee ou non
 */
int isGzip(FILE* tarFile) {

	/* DOC: "These have the fixed values ID1 = 31 (0x1f, \037), ID2 = 139
	 * (0x8b, \213), to identify the file as being in gzip format." */
	return fgetc(tarFile) == 31 && fgetc(tarFile) == 139;
}

/*
 * Charge la librairie zlib dynamiquement
 */
void loaddyn_zlib() {
	compressed.dynopen = dlopen("zlib.so", RTLD_LAZY);

	if(compressed.dynopen == NULL) {
		fprintf(stderr, "[FATAL] Librairie zlib.so pour la compression gzip introuvable\n");
		exit(EXIT_FAILURE);

	} else {
		compressed.gzopen = (gzFile (*)(const char *, const char *)) dlsym(compressed.dynopen, "gzopen");
		compressed.gzread = (int (*)(gzFile, voidp, unsigned)) dlsym(compressed.dynopen, "gzread");
		compressed.gzclose = (int (*)(gzFile)) dlsym(compressed.dynopen, "gzclose");
		compressed.gzeof = (int (*)(gzFile)) dlsym(compressed.dynopen, "gzeof");
		compressed.gzseek = (int (*) (gzFile, int, int)) dlsym(compressed.dynopen, "gzseek");
		compressed.gztell = (int (*) (gzFile)) dlsym(compressed.dynopen, "gztell");
	}
}

/*
 * Ouvrir le fichier, en prenant en compte la compression s'il y a lieu
 * @filename Nom du fichier a ouvrir
 * @return Pointeur vers le fichier ouvert
 */
FILE* pfopen(char* filename) {
	FILE** tarFile;

	/* Allocation dynamique pour pouvoir retourner le pointeur */
	tarFile = malloc(sizeof(FILE *));

	/* Ouverture pour pouvoir consuler le mime-type */
	if((*tarFile = fopen(filename, "r")) == NULL) {
		fprintf(stderr, "[FATAL] Le fichier ne peut etre ouvert.\n");
		exit(EXIT_FAILURE);
	}

	/* Reouverture si c'est un format gzip */
	if(isGzip(*tarFile)) {
		compressed.isGzipped = 1;
		fclose(*tarFile);
		loaddyn_zlib();

		*tarFile = compressed.gzopen(filename, "r");

	/* TODO: elsif avec d'autres formats de compression */

	/* Aucune compression : on laisse ouvert tel quel, mais on revient au
	 * debut du fichier pour le traitement qui suivra */
	} else {
		compressed.isGzipped = 0;
		rewind(*tarFile);
	}

	return *tarFile;
}

/*
 * Ferme un fichier, selon la compression
 * @tarFile Fichier a fermer
 * @return Erreurs
 */
int pfclose(FILE* tarFile) {
	int ret;

	/* Fermeture si c'est un format gzip */
	if(compressed.isGzipped) {
		ret = compressed.gzclose(tarFile);
		dlclose(compressed.dynopen);

	/* TODO: elsif avec d'autres formats de compression */
	
	/* Aucune compression : on ferme le fichier normalement */
	} else
		ret = fclose(tarFile);

	return ret;
}

/*
 * Lit le fichier, selon la compression
 * @ptr Pointeur destine a recuperer les donnnees
 * @size Nombre d'octets a recuperer pour un bloc
 * @nmemb Nombre de blocs
 * @stream Flux du fichier a lire
 * @return Nombre d'items lus
 */
size_t pfread(void* ptr, size_t size, size_t nmemb, FILE *stream) {
	return compressed.isGzipped ? 
		compressed.gzread(stream, ptr, size * nmemb) :
		fread(ptr, size, nmemb, stream);

	/* TODO: elsif avec d'autres formats de compression */
}

/*
 * Indique si le fichier a ete lu dans son integralite, selon la compression
 * @tarFile Fichier a verifier
 * @return A ete lu integralement ou non
 */
int pfeof(FILE* tarFile) {
	return (compressed.isGzipped && compressed.gzeof(tarFile)) ||
		(!compressed.isGzipped && feof(tarFile));

	/* TODO: elsif avec d'autres formats de compression */
}

/*
 * Deplace le curseur de lecture du fichier
 * @tarFile Fichier a modifier
 * @return Erreur ou non
 */
int pfseek(FILE* tarFile, int offset, int whence) {
	return compressed.isGzipped ? 
		compressed.gzseek(tarFile, offset, whence) :
		fseek(tarFile, offset, whence);

	/* TODO: elsif avec d'autres formats de compression */
}

/*
 * Indique l'emplacement courant du curseur
 * @tarFile Fichier a observer
 * @return Emplacement du curseur
 */
int pftell(FILE* tarFile) {
	return compressed.isGzipped ? 
		compressed.gztell(tarFile) :
		ftell(tarFile);

	/* TODO: elsif avec d'autres formats de compression */
}

/*
 * Permet de joindre un thread
 * @return Nombre d'erreurs detectes dans le thread
*/
int join_thread(pthread_t thread) {
	int* status;

	pthread_join(thread, (void **) &status);

	return *status;
}

/*
 * Fonction, avec une gestion des semaphores
 * @args Idem pfread
 * @semaphore Semaphore a respecter
 * @cursor Emplacement courant du curseur sur le fichier
 */
size_t pfread_sem(void* ptr, size_t size, size_t nmemb, FILE *stream, sem_t* semaphore, long* cursor) {
	size_t ret;

	sem_wait(semaphore);
	pfseek(stream, *cursor, SEEK_SET);

	ret = pfread(ptr, size, nmemb, stream);

	*cursor = pftell(stream);
	sem_post(semaphore);

	return ret;
}
