/*
 * ptar - PROJET ESIAL RS 2011/2012
 * Copyright (C) 2011 Julien VAUBOURG / Valentin KELLER
 * Contact <julien@vaubourg.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ptar.h"
#include <sys/time.h>

/**
 **	USAGE : ./% archive.tar
 **	DOC : man 5 tar
 **/

int main(int argc, char** argv) {
	unsigned int i, recordOfNULs, iThread, errors;
	header_posix_ustar header;
        header_for_thread *p_header_t;
	pthread_t* threads;
	struct timeval tic;
	long cursor;
	FILE* tarFile;
	char isData;
	options userOptions;
	sem_t semaphore;

	isData = 1;
	recordOfNULs = 0;
	errors = 0;
	cursor = 0;

	/* Recuperation des options */
	get_options(&userOptions, argc, argv);

	/* Allocation dynamique des tableaux de gestion des threads */
	threads = malloc(userOptions.nbThreads * sizeof(pthread_t));
	threadsAlive = malloc(userOptions.nbThreads * sizeof(unsigned int));

	/* Initialisation des emplacements des threads */
	for(i = 0; i < userOptions.nbThreads; threadsAlive[i++] = FREE_THREAD);

	/* Lancement de l'horloge (si option verbose active) */
	if(userOptions.verbose)
		gettimeofday(&tic, NULL);

	/* Overture de l'archive */
	tarFile = pfopen(userOptions.source);

	/* Initialisation du semaphore */
	sem_init(&semaphore, 0, 1);

	/* Tant que le marqueur de fin de fichier n'a pas ete trouve, et que
	 * la fin de l'archive n'a pas ete rencontree */
	while(isData && !pfeof(tarFile)) {

		/* Lecture du premier enregistrement : soit un entete de fichier
		 * a extraire, soit un enregistrement compose de caracteres NUL */
		pfread_sem(&header, RECORD_SIZE, 1, tarFile, &semaphore, &cursor);

		/* Aucun enregistrement de NULs n'a ete trouve, verification de
		 * celui en cours. DOC : "The end of the archive is indicated by
		 * two records consisting entirely of zero bytes" */
		if(recordOfNULs < 2) {

			isData = 0;

			/* Verifie si l'enregistrement copie est compose de NULs */
			for(
				i = 0;
				i < RECORD_SIZE && !isData;
				isData |= ((char *) &header)[i++]
			);

			/* Si c'est le cas, recordOfNULs passe a 1. La lecture
			 * reviendra ici ensuite, et determinera si un second
			 * enregistrement de NULs est trouve. Si c'est le cas,
			 * recordOfNULs passera a 2, et la lecture de l'archive
			 * cessera. */
			if(!isData)
				recordOfNULs++;

		/* Si un premier enregistrement de NULs a ete trouve, mais que
		 * le suivant n'est pas identique, les deux sont relus, et
		 * le premier accede donc a l'etape suivante. */
		} else
			recordOfNULs = 0;

		/* Extraction du fichier suivant, s'il n'y a pas eu
		 * deux enregistrements de NULs de detectes */
		if(!recordOfNULs) {

			/* Verifie le checksum du header courant, et stoppe
			 * l'extraction s'il est corrompu */
			if(!isValid_checksum(header)) {
				pfclose(tarFile);
				fprintf(stderr, "[FATAL] Somme de controle incorrecte\n");
				exit(EXIT_FAILURE);
			}

			/* Configuration du thread */
			p_header_t = malloc(sizeof(header_for_thread));
			p_header_t->userOptions = &userOptions;
			p_header_t->tarFile = tarFile;
			p_header_t->typeflag = atoi(header.typeflag);
			p_header_t->sem = &semaphore;
			p_header_t->cursor = cursor;

			/* Recuperation des attributs du fichier a extraire */
			get_attributes(&header, &userOptions, &(p_header_t->attr));

			if(!userOptions.list) {

				/* Cherche un emplacement de thread qui est libre ou qui
				 * heberge un thread mort */
				for(
					iThread = 0;
					threadsAlive[iThread % userOptions.nbThreads] == RUNNING_THREAD;
					iThread++
				) { usleep(1); }

				/* Emplacement dispo trouve */
				p_header_t->iThread = iThread % userOptions.nbThreads;

				/* Si l'emplacement correspond a un thread mort, on le
				 * joint avant de lui piquer sa tombe */
				if(threadsAlive[p_header_t->iThread] == DEAD_THREAD)
					errors += join_thread(threads[p_header_t->iThread]);

				/* Lancement d'un thread dedie au fichier courant */
				threadsAlive[p_header_t->iThread] = RUNNING_THREAD;
				pthread_create(&threads[p_header_t->iThread], NULL, (void *) extract_files, (void *) p_header_t);

			/* L'utilisateur ne veut qu'une liste des fichiers */
			} else
				printf("%s\n", p_header_t->attr.filename);

			/* Decalage du curseur de la boucle principale pour atteindre le fichier suivant */
			if(p_header_t->typeflag == FILE_REGULAR)
				cursor += p_header_t->attr.size + (RECORD_SIZE - p_header_t->attr.size % RECORD_SIZE);

			/* Suppression de la structure puisqu'il n'y a pas d'extraction */
			if(userOptions.list)
				free(p_header_t);

		/* Si un enregistrement compose de NULs a ete detecte */
		} else if(recordOfNULs < 2) {

			/* Il s'agit de la verification du second enregistrement :
			 * le premier a genere le recordOfNULs a 1, mais le
			 * second a place le drapeau isData a vrai, ca n'est
			 * donc finalement pas une fin d'archive */
			if(isData) {

				/* La tete de lecture du fichier est renvoyee
				 * deux enregistrements plus tot, pour retraiter
				 * l'enregistrement de NULs sans lui faire
				 * passer le test du suivant. C'est le seul cas
				 * ou le meme enregistrement est lu deux fois,
				 * mais c'est un cas extremement rare, peut-etre
				 * meme impossible a rencontrer, malgres la doc. */
				cursor += -1 * RECORD_SIZE;

				recordOfNULs++;

			/* Un premier enregistrement de NULs a ete trouve : la
			 * lecture est relancee pour determiner la valeur du
			 * suivant */
			} else
				isData = 0;
		}
	}

	/* Pour que le programme soit vraiment durable, on attend la fin de
	 * chacun des threads encore en activite */
	for(i = 0; i < userOptions.nbThreads; i++) {
		if(threadsAlive[i] != FREE_THREAD)
			errors += join_thread(threads[i]);
	}

	pfclose(tarFile);
	sem_destroy(&semaphore);
	free(threads);
	free(threadsAlive);

	/* Arret de l'horloge (si mode verbose active) */
	if(userOptions.verbose)
		stopTimer(&tic);

	/* Fin du programme */
	if(errors > 0) {
		fprintf(
			stderr,
			"[ERREUR] %s : Il y a eu %d erreur%s durant l'extraction\n",
			userOptions.source,
			errors,
			errors > 1 ? "s" : ""
		);

		return EXIT_FAILURE;
	} else
		return EXIT_SUCCESS;
}
